// Cytosim was created by Francois Nedelec. Copyright 2021 Cambridge University
/**
 @file
 @brief  Dimensionality of the simulation
 Note that the setting in this file can be overruled
 by providing -DDIM=3 to the compiler command.
 */

#ifndef DIM

/**
 DIM defines the number of dimensions in space:
 it must be 1, 2 or 3
 */
#define DIM 3

#endif
